from typing import TYPE_CHECKING, Any, List, Optional, Union
from ain.account import Account, Accounts
from ain.types import (
    TransactionBody,
    ValueOnlyTransactionInput,
    ECDSASignature,
)
from ain.utils import (
    privateToAddress,
    toBytes,
    toChecksumAddress,
    bytesToHex,
    pubToAddress,
    ecSignMessage,
    ecSignTransaction,
    ecRecoverPub,
    ecVerifySig,
)
from ain.utils.v3keystore import V3Keystore, V3KeystoreOptions

if TYPE_CHECKING:
    from ain.ain import Ain

class Wallet:
    """Class for the AIN Blockchain wallet."""

    defaultAccount: Optional[Account]
    """The default account of the wallet."""
    accounts: Accounts
    """The set of the accounts. Keys are the addresses of the accounts."""
    ain: "Ain"
    """The `Ain` instance."""
    chainId: int
    """The chain ID of the provider."""

    def __init__(self, ain: "Ain", chainId: int):
        self.defaultAccount = None
        self.accounts: Accounts = {}
        self.ain = ain
        self.chainId = chainId

    @property
    def length(self) -> int:
        return len(self.accounts)

    def getPublicKey(self, address: str) -> str:
        """Gets a full public key of the given address.

        Args:
            address(str): The address of the account.

        Returns:
            str: The full public key of the given address.
                If such address is not in the wallet, returns empty string.
        """
        checksummed = toChecksumAddress(address)
        if checksummed not in self.accounts:
            return ''
        return self.accounts[checksummed].public_key

    def create(self, numberOfAccounts: int):
        """Creates `numberOfAccounts` new accounts, and adds them to the wallet.

        Args:
            numberOfAccounts(int): The number of account that you want to create.
                If this is not positive, raises `ValueError`.
        """
        if numberOfAccounts <= 0:
            raise ValueError("numberOfAccounts should be greater than 0.")
        
        newAccounts: List[str] = []
        for i in range(numberOfAccounts):
            account = Account.create()
            self.accounts[account.address] = account
            newAccounts.append(account.address)

        return newAccounts
    
    def isAdded(self, address: str) -> bool:
        """Returns whether the address has already been added to the wallet.

        Args:
            address (str): The address of the account.

        Returns:
            bool: `True`, if the address has already been added to the wallet,
                `False`, if not.
        """
        return address in self.accounts

    def add(self, privateKey: Union[str, bytes]) -> str:
        """Adds a new account to the wallet, from the given private key.

        Args:
            privateKey (Union[str, bytes]): The private key of the new account.
        
        Returns:
            str: The address of the new account.
        """
        newAccount = Account(privateKey)
        self.accounts[newAccount.address] = newAccount
        return newAccount.address

    def addAndSetDefaultAccount(self, privateKey: Union[str, bytes]) -> str:
        """Adds a new account to the wallet, from the given private key.
        Then, sets the new account as the default account.

        Args:
            privateKey (Union[str, bytes]): The private key of the new account.
        
        Returns:
            str: The address of the new account.
        """
        address = self.add(privateKey)
        self.setDefaultAccount(address)
        return address

    def addFromHDWallet(self, mnemonic: str, index: int = 0) -> str:
        """Adds an account to the wallet, from a seed phrase.
        Only the account at the given index will be added.

        Args:
            mnemonic (str): The mnemonic of account.
            index (int): The index of account. Defaults to 0.
        
        Returns:
            str: The address of the new account.
        """
        account = Account.fromMnemonic(mnemonic, index)
        self.accounts[account.address] = account
        return account.address

    def addFromV3Keystore(self, v3Keystore: Union[V3Keystore, str], password: str) -> str:
        """Adds an account to the wallet, from the v3 keystore.

        Args:
            mnemonic (Union[V3Keystore, str]): The v3 keystore of account.
            index (int): The password of your v3 keystore.
        
        Returns:
            str: The address of the new account.
        """
        if type(v3Keystore) is V3Keystore:
            privateKey = v3Keystore.toPrivateKey(password)
        elif type(v3Keystore) is str:
            privateKey = V3Keystore.fromJSON(v3Keystore).toPrivateKey(password)
        else:
            raise TypeError("v3Keystore has invalid type")
        self.add(privateKey)
        return privateToAddress(privateKey)
    
    def remove(self, address: str):
        """Removes the account from the wallet.
        If such account doesn't exist, raises `ValueError`.

        Args:
            address (str): The address of the account.
        """
        addressToRemove = toChecksumAddress(address)
        if addressToRemove not in self.accounts:
            raise ValueError("Can't find account to remove")
        del self.accounts[addressToRemove]
        if self.defaultAccount is not None and self.defaultAccount.address == addressToRemove:
            self.removeDefaultAccount()

    def setDefaultAccount(self, address: str):
        """Sets the default AIN Blockchain account address as `address`.
        If such `address` is not added to the wallet, raises `ValueError`.

        Args:
            address (str): The address of the account.
        """
        checksumed = toChecksumAddress(address)
        if checksumed not in self.accounts:
            raise ValueError("Add the account first before setting it to defaultAccount.")
        self.defaultAccount = self.accounts[checksumed]

    def removeDefaultAccount(self):
        """Removes a default account from the wallet, and sets it to `None`."""
        self.defaultAccount = None

    def clear(self):
        """Clears the wallet, remove all account information."""
        self.accounts.clear()
        self.removeDefaultAccount()

    def getImpliedAddress(self, address: str = None) -> str:
        """Gets an implied address.

        Args:
            address (str, Optional): The address of the account.
                If `address` is not given, returns the default account address. Defaults to `None`.
                If `address` is not given and default account is not set, raises `ValueError`.
                If such `address` is not added to the wallet, raises `ValueError`.

        Returns:
            str: The implied address.
        """
        if address is not None:
            filteredAddress = address
        elif self.defaultAccount is not None:
            filteredAddress = self.defaultAccount.address
        else:
            raise ValueError("You need to specify the address or set defaultAccount.")
        checksummed = toChecksumAddress(filteredAddress)
        if checksummed not in self.accounts:
            raise ValueError("The address you specified is not added in your wallet. Try adding it first.")
        return checksummed

    async def getBalance(self, address: str = None) -> int:
        """Gets an AIN balance of the given account.

        Args:
            address (str, Optional): The address of the account.
                If `address` is `None`, returns the balance of the default account address. Defaults to `None`.
                If `address` is `None` and default account is not set, raises `ValueError`.

        Returns:
            str: The AIN balance of the given account.
        """
        if address is None:
            addr = self.getImpliedAddress()
        else:
            addr = toChecksumAddress(address)
        return await self.ain.db.ref(f"/accounts/{addr}/balance").getValue()

    async def transfer(self, toAddress: str, value: int, fromAddress: str = None, nonce: int = None, gas_price: int = None):
        """Sends a transfer transaction to the network.

        Args:
            toAddress (str): The AIN blockchain address that wants to transfer AIN to.
            value (int): The amount of the transferring AIN.
            fromAddress (str, Optional): The AIN blockchain address that wants to transfer AIN from.
                Defaults to `None`, transfer from the default account of the current wallet.
            nonce (int, Optional): The nonce of the transfer transaction.
                Defaults to `None`.
            gas_price (int, Optional): The gas price of the transfer transaction.
                Defaults to `None`.
        
        Returns:
            The transaction result.
        """
        fromAddr = self.getImpliedAddress(fromAddress)
        toAddr = toChecksumAddress(toAddress)
        transferRef = self.ain.db.ref(f"/transfer/{fromAddr}/{toAddr}").push()
        return await transferRef.setValue(
            ValueOnlyTransactionInput(
                ref="/value",
                address=fromAddress,
                value=value,
                nonce=nonce,
                gas_price=gas_price,
            )
        )

    def sign(self, data: str, address: str = None) -> str:
        """Signs a string data with the private key of the given address.

        Args:
            data (str): The string data.
            address (str, Optional): The AIN blockchain address that wants to use for signing.
                Defaults to `None`, use the default account of the current wallet.
        
        Returns:
            str: The signature of the data.
        """
        addr = self.getImpliedAddress(address)
        privateKeyBytes = bytes.fromhex(self.accounts[addr].private_key)
        return ecSignMessage(data, privateKeyBytes, self.chainId)
    
    def signTransaction(self, tx: TransactionBody, address: str = None) -> str:
        """Signs a transaction data with the private key of the given address.

        Args:
            tx (TransactionBody): The transaction data.
            address (str, Optional): The AIN blockchain address that wants to use for signing.
                Defaults to `None`, use the default account of the current wallet.
        
        Returns:
            str: The signature of the transaction.
        """
        addr = self.getImpliedAddress(address)
        privateKeyBytes = bytes.fromhex(self.accounts[addr].private_key)
        return ecSignTransaction(tx, privateKeyBytes, self.chainId)

    def getHashStrFromSig(self, signature: str) -> str:
        """Gets a hash from the signature.

        Args:
            signature (str): The signature.

        Returns:
            The hex prefixed hash from the signature.
        """
        sigBytes = toBytes(signature)
        lenHash = len(sigBytes) - 65
        hashedData = sigBytes[0:lenHash]
        return "0x" + hashedData.hex()

    def recover(self, signature: str) -> str:
        """Recovers an address of the account that was used to create the signature.

        Args:
            signature (str): The signature.
        
        Returns:
            The address of the account that was used to create the signature.
        """
        sigBytes = toBytes(signature)
        lenHash = len(sigBytes) - 65
        hashedData = sigBytes[0:lenHash]
        sig = ECDSASignature.fromSignature(sigBytes[lenHash:])
        return toChecksumAddress(bytesToHex(pubToAddress(ecRecoverPub(hashedData, sig, self.chainId)[1:])))

    def verifySignature(self, data: Any, signature: str, address: str) -> bool:
        """Verifies the given signature.
        
        Args:
            data: The data that has made the signature.
            signature (str): The signature of the given data.
            address (str): The address of who created the signature.
        
        Returns:
            bool: `True`, if the signature is valid.
                `False`, if not.
        """
        return ecVerifySig(data, signature, address, self.chainId)

    def toV3Keystore(
        self,
        password: str,
        options: V3KeystoreOptions = V3KeystoreOptions()
    ) -> List[V3Keystore]:
        """Saves the accounts in the wallet as v3 Keystores, locking them with the password.

        Args:
            password (str): The password of the v3 keystores.
            options (V3KeystoreOptions): The options for the v3 keystores.
                Defaults to no options.
        
        Returns:
            List[V3Keystore]: The list of the v3 Keystores.
        """
        ret = []
        for address in self.accounts:
            ret.append(self.accountToV3Keystore(address, password, options))
        return ret
    
    def accountToV3Keystore(
        self,
        address: str,
        password: str,
        options: V3KeystoreOptions = V3KeystoreOptions()
    ) -> V3Keystore:
        """Converts an account into a v3 Keystore and encrypts it with a password.

        Args:
            address: The AIN blockchain address.
            password: The password of the v3 keystore.
            options (V3KeystoreOptions): The options for the v3 keystore.
                Defaults to no options.
                
        Returns:
            V3Keystore: The v3 Keystore of the account.
        """
        if not self.isAdded(address):
            raise ValueError("No such address exists in the wallet")
        privateKey = self.accounts[address].private_key
        return V3Keystore.fromPrivateKey(privateKey, password, options)


import sys
from typing import Any, Optional, List, Union

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

class ECDSASignature:
    """Class for the ECDSA signature."""

    r: bytes
    s: bytes
    v: int

    def __init__(self, r: bytes, s: bytes, v: int):
        if len(r) != 32:
            raise ValueError("Invalid signature length: r")
        if len(s) != 32:
            raise ValueError("Invalid signature length: s")
        if v < 0 or v > 255:
            raise ValueError("Invalid signature: v")

        self.r = r
        self.s = s
        self.v = v

    @classmethod
    def fromSignature(cls, signature: bytes):
        """Converts signature format of the `eth_sign` RPC method to signature parameters.

        Args:
            signature: The `eth_sign` signature.
        """
        if len(signature) != 65:
            raise ValueError("Invalid signature length: r")

        r = signature[0:32]
        s = signature[32:64]
        v = signature[64]
        return cls(r, s, v)

class ECIESEncrypted:
    """Class for the ECIES encrypted ciphertext."""

    iv: bytes
    ephemPublicKey: bytes
    ciphertext: bytes
    mac: bytes

    def __init__(
        self,
        iv: bytes,
        ephemPublicKey: bytes,
        ciphertext: bytes,
        mac: bytes,
    ):
        self.iv = iv
        self.ephemPublicKey = ephemPublicKey
        self.ciphertext = ciphertext
        self.mac = mac

    def __str__(self):
        return "\n".join(
            (
                "{",
                f"  iv: '{self.iv.hex()}',",
                f"  ephemPublicKey: '{self.ephemPublicKey.hex()}',",
                f"  ciphertext: '{self.ciphertext.hex()}',",
                f"  mac: '{self.mac.hex()}',",
                "}",
            )
        )

SetMultiOperationType = Literal["SET"]

GetMultiOperationType = Literal["GET"]

SetOperationType = Literal["SET_VALUE", "INC_VALUE", "DEC_VALUE", "SET_RULE", "SET_OWNER", "SET_FUNCTION"]

GetOperationType = Literal["GET_VALUE", "GET_RULE", "GET_OWNER", "GET_FUNCTION"]

OwnerPermission = Literal["branch_owner", "write_function", "write_owner", "write_rule"]

class AinOptions():
    """Class for the AIN blockchain options."""

    rawResultMode: Optional[bool]
    """The flag of the using transaction raw result mode. If `True`, the raw result will return."""

    def __init__(
        self,
        rawResultMode: bool = None,
    ):
        if rawResultMode is not None:
            self.rawResultMode = rawResultMode

class GetOptions():
    """Class for the get transaction options."""

    is_global: Optional[bool]
    """The flag of the using global reference path. If `True`, the given ref will be interpreted as a global path."""
    is_final: Optional[bool]
    """The flag of the getting finalized result. If `True`, the result will return finalized."""
    is_shallow: Optional[bool]
    """The flag of the getting shallow result. If `True`, the result will return only the keys of the children."""
    include_version: Optional[bool]
    """The flag of the including version. If `True`, the result will include state versions."""
    include_tree_info: Optional[bool]
    """The flag of the including tree info. If `True`, the result will include additional state tree information."""
    include_proof: Optional[bool]
    """The flag of the including proof hashes. If `True`, the result will include proof hashes."""

    def __init__(
        self,
        is_global: bool = None,
        is_final: bool = None,
        is_shallow: bool = None,
        include_version: bool = None,
        include_tree_info: bool = None,
        include_proof: bool = None,
    ):
        if is_global is not None:
            self.is_global = is_global
        if is_final is not None:
            self.is_final = is_final
        if is_shallow is not None:
            self.is_shallow = is_shallow
        if include_version is not None:
            self.include_version = include_version
        if include_tree_info is not None:
            self.include_tree_info = include_tree_info
        if include_proof is not None:
            self.include_proof = include_proof

class SetOperation():
    """Class for the set operation."""

    type: SetOperationType
    """The type of the set operation."""
    ref: str
    """The reference of the path."""
    value: Optional[Any]
    """The value that you want to set."""
    is_global: Optional[bool]
    """The flag of the using global reference path. If `True`, the given ref will be interpreted as a global path."""

    def __init__(
        self,
        type: SetOperationType,
        ref: str,
        value: Any = None,
        is_global: bool = None,
    ):
        self.type = type
        self.ref = ref
        if value is not None:
            self.value = value
        if is_global is not None:
            self.is_global = is_global

class SetMultiOperation():
    """Class for the multiple set operation."""

    type: SetMultiOperationType
    """The type of the multiple set operation."""
    op_list: List[SetOperation]
    """The list of the set operations."""

    def __init__(
        self,
        type: SetMultiOperationType,
        op_list: List[SetOperation]
    ):
        self.type = type
        self.op_list = op_list

class GetOperation():
    """Class for the get operation."""

    type: GetOperationType
    """The type of the get operation."""
    ref: Optional[str]
    """The path that you want to make transaction."""
    is_global: Optional[bool]
    """The flag of the using global reference path. If `True`, the given ref will be interpreted as a global path."""

    def __init__(
        self,
        type: GetOperationType,
        ref: str = None,
        is_global: bool = None,
    ):
        self.type = type
        if ref is not None:
            self.ref = ref
        if is_global is not None:
            self.is_global = is_global

class GetMultiOperation():
    """Class for the multiple get operation."""

    type: GetMultiOperationType
    """The type of the multiple get operation."""
    op_list: List[GetOperation]
    """The list of the get operations."""

    def __init__(
        self,
        type: GetMultiOperationType,
        op_list: List[GetOperation]
    ):
        self.type = type
        self.op_list = op_list

class TransactionBodyBase:
    """Class for the transaction body base."""

    parent_tx_hash: Optional[str]
    """The hash of the parent transaction. If transaction is not nested, this should not be given."""
    operation: Union[SetOperation, SetMultiOperation]
    """The set operation."""

class ValueOnlyTransactionBodyBase:
    """Class for the value only transaction body base."""

    parent_tx_hash: Optional[str]
    """The hash of the parent transaction. If transaction is not nested, this should not be given."""
    value: Optional[Any]
    """The value that you want to set."""
    ref: Optional[str]
    """The path that you want to make transaction."""
    is_global: Optional[bool]
    """The flag of the using global reference path. If `True`, the given ref will be interpreted as a global path."""

class TransactionInputBase:
    """Class for the transaction input base."""

    nonce: Optional[int]
    """The nonce of the transaction."""
    address: Optional[str]
    """The AIN blockchain address of the sender."""
    timestamp: Optional[int]
    """The unix time in milliseconds when the transaction has been created."""
    gas_price: Optional[int]
    """The gas price of the transaction."""

class TransactionBody(TransactionBodyBase):
    """Class for the transaction body."""

    nonce: int
    """The nonce of the transaction."""
    timestamp: int
    """The unix time in milliseconds when the transaction has been created."""
    gas_price: Optional[int]
    """The gas price of the transaction."""
    billing: Optional[str]
    """The billing app name and the billing ID. Must be formatted as `<app_name>|<billing_id>`"""

    def __init__(
        self,
        operation: Union[SetOperation, SetMultiOperation],
        nonce: int,
        timestamp: int,
        gas_price: int = None,
        billing: str = None,
        parent_tx_hash: str = None,
    ):
        self.operation = operation
        self.nonce = nonce
        self.timestamp = timestamp
        if gas_price is not None:
            self.gas_price = gas_price
        if billing is not None:
            self.billing = billing
        if parent_tx_hash is not None:
            self.parent_tx_hash = parent_tx_hash


class TransactionInput(TransactionBodyBase, TransactionInputBase):
    """Class for the transaction input."""

    def __init__(
        self,
        operation: Union[SetOperation, SetMultiOperation],
        parent_tx_hash: str = None,
        nonce: int = None,
        address: str = None,
        timestamp: int = None,
        gas_price: int = None,
    ):
        self.operation = operation
        if parent_tx_hash is not None:
            self.parent_tx_hash = parent_tx_hash
        if nonce is not None:
            self.nonce = nonce
        if address is not None:
            self.address = address
        if timestamp is not None:
            self.timestamp = timestamp
        if gas_price is not None:
            self.gas_price = gas_price

class ValueOnlyTransactionInput(ValueOnlyTransactionBodyBase, TransactionInputBase):
    """Class for the value only transaction input."""

    def __init__(
        self,
        parent_tx_hash: str = None,
        value: Any = None,
        ref: str = None,
        is_global: bool = None,
        nonce: int = None,
        address: str = None,
        timestamp: int = None,
        gas_price: int = None,
    ):
        if parent_tx_hash is not None:
            self.parent_tx_hash = parent_tx_hash
        if value is not None:
            self.value = value
        if ref is not None:
            self.ref = ref
        if is_global is not None:
            self.is_global = is_global
        if nonce is not None:
            self.nonce = nonce
        if address is not None:
            self.address = address
        if timestamp is not None:
            self.timestamp = timestamp
        if gas_price is not None:
            self.gas_price = gas_price

class SetMultiTransactionInput(TransactionInputBase):
    """Class for the multiple set transaction input."""

    parent_tx_hash: Optional[str]
    """The hash of the parent transaction. If transaction is not nested, this should not be given."""
    op_list: List[SetOperation]
    """The list of the set operations."""

    def __init__(
        self,
        op_list: List[SetOperation],
        parent_tx_hash: str = None,
        nonce: int = None,
        address: str = None,
        timestamp: int = None,
        gas_price: int = None,
    ):
        self.op_list = op_list
        if parent_tx_hash is not None:
            self.parent_tx_hash = parent_tx_hash
        if nonce is not None:
            self.nonce = nonce
        if address is not None:
            self.address = address
        if timestamp is not None:
            self.timestamp = timestamp
        if gas_price is not None:
            self.gas_price = gas_price

class Transaction():
    """Class for the transaction."""

    tx_body: TransactionBody
    """The transaction body."""
    signature: str
    """The signature of the transaction."""
    hash: str
    """The transaction hash."""
    address: str
    """The AIN blockchain address of the sender."""

class EvalRuleInput:
    """Class for the eval rule input."""

    value: Any
    """The value that you want to set."""
    ref: Optional[str]
    """The path that you want to make transaction."""
    address: Optional[str]
    """The AIN blockchain address of the sender."""
    timestamp: Optional[int]
    """The unix time in milliseconds when the transaction has been created."""
    is_global: Optional[bool]
    """The flag of the using global reference path. If `True`, the given ref will be interpreted as a global path."""

    def __init__(
        self,
        value: Any = None,
        ref: str = None,
        address: str = None,
        timestamp: int = None,
        is_global: bool = None,
    ):
        if value is not None:
            self.value = value
        if ref is not None:
            self.ref = ref
        if address is not None:
            self.address = address
        if timestamp is not None:
            self.timestamp = timestamp
        if is_global is not None:
            self.is_global = is_global

class EvalOwnerInput:
    """Class for the eval owner input."""

    permission: OwnerPermission
    """The type of the permission."""
    ref: Optional[str]
    """The path that you want to make transaction."""
    address: Optional[str]
    """The AIN blockchain address of the sender."""
    is_global: Optional[bool]
    """The flag of the using global reference path. If `True`, the given ref will be interpreted as a global path."""

    def __init__(
        self,
        permission: OwnerPermission,
        ref: str = None,
        address: str = None,
        is_global: bool = None,
    ):
        self.permission = permission
        if ref is not None:
            self.ref = ref
        if address is not None:
            self.address = address
        if is_global is not None:
            self.is_global = is_global

class MatchInput:
    """Class for the match input."""

    ref: Optional[str]
    """The path that you want to make transaction."""
    is_global: Optional[bool]
    """The flag of the using global reference path. If `True`, the given ref will be interpreted as a global path."""

    def __init__(
        self,
        ref: str = None,
        is_global: bool = None,
    ):
        if ref is not None:
            self.ref = ref
        if is_global is not None:
            self.is_global = is_global



from typing import Dict, Union
from secrets import token_bytes
from ain.utils import (
    privateToPublic,
    privateToAddress,
    mnemonicToPrivatekey,
    keccak
)

class Account:
    """Class for the account on the AIN Blockchain.
    
    Args:
        privateKey (Union[bytes, str]): The private key of the account.
    """

    private_key: str
    """The private key of the account."""
    public_key: str
    """The public key of the account."""
    address: str
    """The address of the account on the AIN Blockchain"""
    
    def __init__(self, privateKey: Union[bytes, str] = None):
        if type(privateKey) is bytes:
            privateKeyBytes = privateKey
        elif type(privateKey) is str:
            privateKeyBytes = bytes.fromhex(privateKey)
        else:
            raise TypeError('privateKey has invalid type')

        self.private_key = privateKeyBytes.hex()
        self.public_key = privateToPublic(privateKeyBytes).hex()
        self.address = privateToAddress(privateKeyBytes)

    @classmethod
    def fromMnemonic(cls, mnemonic: str, index: int = 0):
        """Inits an `Account` with the given mnemonic.

        Args:
            mnemonic (str): The mnemonic of account.
            index (int): The index of account. Defaults to 0.
        """
        return cls(mnemonicToPrivatekey(mnemonic, index))

    @classmethod
    def create(cls, entropy: str = None):
        """Inits an `Account` with the given entropy.

        Args:
            entropy (str): The entropy of account. Defaults to None.
        """
        entropyBytes = token_bytes(32)
        if entropy is not None:
            entropyBytes = bytes(entropy, "utf-8")

        innerHex = keccak(token_bytes(32) + entropyBytes)
        middleHex = token_bytes(32) + innerHex + token_bytes(32)
        return cls(keccak(middleHex))

    def __str__(self):
        return "\n".join(
            (
                "{",
                f"  address: '{self.address}'",
                f"  private: '{self.private_key}'",
                f"  public: '{self.public_key}'",
                "}",
            )
        )

Accounts = Dict[str, Account]



class BlockchainError(Exception):
    """Class for the custom error of the blockchain API result."""

    code: int
    """The code of an error."""
    message: str
    """The message of an error."""

    def __init__(self, code: int, message: str):
        self.code = code
        self.message = message
        super().__init__(message)

    def __str__(self):
        return f"{str(self.code)}: {str(self.message)}"



import re
import json
import time
from typing import Any, Union
from secrets import token_bytes
from Crypto.Cipher import AES
from Crypto.Hash import keccak as _keccak
from Crypto.Hash import HMAC, SHA256, SHA512
from Crypto.Util.Padding import pad, unpad
from coincurve import PrivateKey, PublicKey
from coincurve.ecdsa import deserialize_recoverable, recoverable_convert, cdata_to_der
from coincurve.utils import validate_secret
from mnemonic import Mnemonic
# TODO(kriii): Need to replace bip32 package or wait for the type hint.
from bip32 import BIP32 # type: ignore
from ain.types import ECDSASignature, ECIESEncrypted, TransactionBody

def encodeVarInt(number: int) -> bytes:
    """Encodes the number as bitcoin variable length integer

    Args:
        number (int): A number that you want to encode.

    Returns:
        bytes: The encoded number as bitcoin variable length integer.
    """
    MAX_SAFE_INTEGER = 9007199254740991
    if number < 0 or number > MAX_SAFE_INTEGER:
        raise ValueError("Number out of range")

    if number < 0xfd:
        byte = 1
    elif number <= 0xffff:
        number = (number << 8) + 0xfd
        byte = 3
    elif number <= 0xffffffff:
        number = (number << 8) + 0xfe
        byte = 5
    else:
        number = (number << 8) + 0xff
        byte = 9

    return number.to_bytes(byte, "little")

SIGNED_MESSAGE_PREFIX = "AINetwork Signed Message:\n"
SIGNED_MESSAGE_PREFIX_BYTES = bytes(SIGNED_MESSAGE_PREFIX, "utf-8")
SIGNED_MESSAGE_PREFIX_LENGTH = encodeVarInt(len(SIGNED_MESSAGE_PREFIX))
AIN_HD_DERIVATION_PATH = "m/44'/412'/0'/0/"

def getTimestamp() -> int:
    """Gets the current unix timestamp.
    
    Returns:
        int: The current unix timestamp.
    """
    return int(time.time() * 1000)

def addHexPrefix(input: str) -> str:
    """Adds the hex prefix('0x') to the given string if it does not hex prefixed.
    
    Args:
        input (str): The string that you want to add prefix.
    
    Returns:
        str: The hex prefixed string.
    """
    return input if isHexPrefixed(input) else ("0x" + input)

def bytesToHex(input: bytes) -> str:
    """Converts the given bytes into the hex prefixed hex string.

    Args:
        input (bytes): The bytes that you want to convert into hex.

    Returns:
        str: The hex converted bytes.
    """
    return "0x" + input.hex()

def isValidAddress(address: str) -> bool:
    """Checks if the given string is a valid address.

    Args:
        address (str): The address that you want to check.

    Returns:
        bool: `True`, if the `address` is valid address.
            `False`, if not.
    """
    pattern = re.compile("0x[0-9a-fA-F]{40}")
    return pattern.fullmatch(address) is not None

def isHexString(input: str) -> bool:
    """Checks if the given string is a hex string.
    
    Args:
        input (str): The string that you want to check.
    
    Returns:
        bool: `True`, if the `input` is hex string.
            `False`, if not.
    """
    pattern = re.compile("0x[0-9a-fA-F]*")
    return pattern.fullmatch(input) is not None

def isHexPrefixed(input: str) -> bool:
    """Checks if the given string is prefixed with `0x`.
        
    Args:
        input (str): The string that you want to check.
    
    Returns:
        bool: `True`, if the `input` is hex prefixed.
            `False`, if not.
    """
    return input.startswith("0x")

def stripHexPrefix(input: str) -> str:
    """Strips `0x` from the given string, if that is hex prefixed.
    
    Args:
        input (str): The string that you want to strip.
    
    Returns:
        str: The hex prefix stripped string, if `input` is hex prefixed.
            `input` itself, if it not hex prefixed.
    """
    return input[2:] if isHexPrefixed(input) else input

def padToEven(input: str) -> str:
    """Pads the leading zero into the given string to have an even length.

    Args:
        input (str): The string that you want to pad.
    
    Returns:
        str: The leading zero padded string, if `input` is odd lengthed.
            `input` itself, if it is even lengthed.
    """
    return input if len(input) % 2 == 0 else ("0" + input)

def toBytes(input: Any) -> bytes:
    """Attempts to convert a input into a bytes,
    This function is equivalent to `toBuffer` from `@ainblockchain/ain-util`.

    Args:
        input: The input that you want to convert.
            It supports bytes, list(of ints), string, int, and None.
            If `input` is unsupported type, raises `TypeError`.

    Returns:
        bytes: The converted bytes.
    """
    if type(input) is bytes:
        pass
    elif type(input) is list:
        input = bytes(input)
    elif type(input) is str:
        if isHexString(input):
            input = bytes.fromhex(padToEven(stripHexPrefix(input)))
        else:
            input = bytes(input, "utf-8")
    elif type(input) is int:
        input = bytes.fromhex(padToEven(stripHexPrefix(hex(input))))
    elif input is None:
        input = bytes(0)
    else:
        raise TypeError("Invalid type")
    return input

def toChecksumAddress(address: str) -> str:
    """Returns a checksummed address.

    Args:
        address (str): The address that you want to convert.
            If `address` isn't valid, raises `ValueError`.

    Returns:
        str: The checksummed address.
    """
    if not isValidAddress(address):
        raise ValueError("Invalid address")

    address = stripHexPrefix(address).lower()
    hash = keccak(address).hex()
    ret = "0x"

    for i in range(len(address)):
        if int(hash[i], base=16) >= 8:
            ret += address[i].upper()
        else:
            ret += address[i]

    return ret

def keccak(input: Any, bits: int = 256) -> bytes:
    """Creates the Keccak hash of the input.

    Args:
        input: The input that you want to create hash.
            Note that `input` will be converted with `toBytes` method.
            If `input` is unsupported type, raises `TypeError`.
        bits (int): The size of the hash, in (224, 456, 384, 512).
            Defaults to 256.
        
    Returns:
        bytes: The Keccak hash of the input.
    """
    inputBytes = toBytes(input)
    k = _keccak.new(digest_bits=bits)
    k.update(inputBytes)
    return k.digest()

def hashTransaction(transaction: Union[TransactionBody, str]) -> bytes:
    """Creates the Keccak-256 hash of the transaction body.

    Args:
        transaction (Union[TransactionBody, str]):
            The transaction that you want to create hash.

    Returns:
        bytes: The Keccak hash of the transaction.
    """
    if type(transaction) is TransactionBody:
        transaction = json.dumps(
            transaction.__dict__,
            default=lambda o: o.__dict__,
            separators=(",", ":"),
            sort_keys=True
        )
    return keccak(keccak(transaction))

def hashMessage(message: Any) -> bytes:
    """Creates the bitcoin's varint encoding of Keccak-256 hash of the message,
    prefixed with the header `AINetwork Signed Message:`.

    Args:
        message: The message that you want to create hash.
            Note that `message` should be convertible with :meth:`toBytes`.
            If `message` is unsupported type, raises `TypeError`.
    
    Returns:
        bytes: The bitcoin's varint encoding of Keccak-256 hash of the message.
    """
    msgBytes = toBytes(message)
    msgLenBytes = encodeVarInt(len(msgBytes))
    dataBytes = bytes(
        bytearray(SIGNED_MESSAGE_PREFIX_LENGTH)
        + bytearray(SIGNED_MESSAGE_PREFIX_BYTES)
        + bytearray(msgLenBytes)
        + bytearray(msgBytes)
    )
    return keccak(keccak(dataBytes))

def ecSignHash(msgHash: bytes, privateKey: bytes, chainId: int = 0) -> ECDSASignature:
    """Returns the ECDSA signature of a message hash.

    Args:
        msgHash (bytes): The message hash.
        privateKey (bytes): The private key for the signature.
        chainId (int): The chain ID of the provider. Defaults to 0.

    Returns:
        ECDSASignature: The object of the ECDSA signature.
    """
    pbK = PrivateKey(privateKey)
    sig = pbK.sign_recoverable(msgHash, hasher=None)
    r = sig[0:32]
    s = sig[32:64]
    v = sig[64] + (27 if chainId == 0 else chainId * 2 + 35)
    return ECDSASignature(r, s, v)

def ecSignMessage(message: Any, privateKey: bytes, chainId: int = 0) -> str:
    """Signs a message with a private key and returns a string signature.

    Args:
        message: The message that you want to sign.
            Note that `message` should be convertible with :meth:`toBytes`.
            If `message` is unsupported type, raises `TypeError`.
        privateKey (bytes): The private key for the signature.
        chainId (int): The chain ID of the provider. Defaults to 0.
    
    Returns:
        str: The hex prefixed string signature.
    """
    hashedMsg = hashMessage(message)
    signature = ecSignHash(hashedMsg, privateKey, chainId)
    return "0x" + hashedMsg.hex() + signature.r.hex() + signature.s.hex() + toBytes(signature.v).hex()

def ecSignTransaction(txData: TransactionBody, privateKey: bytes, chainId: int = 0) -> str:
    """Signs a transaction body with a private key and returns a string signature.

    Args:
        txData (TransactionBody): The transaction that you want to sign.
        privateKey (bytes): The private key for the signature.
        chainId (int): The chain ID of the provider. Defaults to 0.

    Returns:
        str: The hex prefixed string signature.
    """
    hashedTx = hashTransaction(txData)
    signature = ecSignHash(hashedTx, privateKey, chainId)
    return "0x" + hashedTx.hex() + signature.r.hex() + signature.s.hex() + toBytes(signature.v).hex()

def ecRecoverPub(msgHash: bytes, signature: ECDSASignature, chainId: int = 0) -> bytes:
    """Recovers the public key from ECDSA signature.

    Args:
        msgHash (bytes): The hash of the message.
        signature (ECDSASignature): The object of the ECDSA signature.
        chainId (int): The chain ID of the provider. Defaults to 0.

    Returns:
        str: The recovered public key.
    """
    recovery = signature.v - (27 if chainId == 0 else chainId * 2 + 35)
    if recovery != 0 and recovery != 1:
        raise ValueError("Invalid signature v value")
    cSig = bytes(bytearray(signature.r) + bytearray(signature.s) + bytearray([recovery]))
    senderPubKey = PublicKey.from_signature_and_message(cSig, msgHash, hasher=None)
    nSig = cdata_to_der(recoverable_convert(deserialize_recoverable(cSig)))
    if not senderPubKey.verify(nSig, msgHash, hasher=None):
        raise ValueError('Public key verify failed')
    return senderPubKey.format(False)

def ecVerifySig(data: Any, signature: str, address: str, chainId: int = 0) -> bool:
    """Verifies if the signature is valid.

    Args:
        data: The message or transaction that you want to verify.
        signature (str): The hex prefixed string signature.
        address (str): The address of the signature.
        chainId (int): The chain ID of the provider. Defaults to 0.

    Returns:
        bool: `True`, if the given signature is a valid signature.
            `False`, if not.
    """
    sigBytes = toBytes(signature)
    lenHash = len(sigBytes) - 65
    hashedData = sigBytes[:lenHash]
    if type(data) is TransactionBody:
        if hashedData != hashTransaction(data):
            return False
    else:
        if hashedData != hashMessage(data):
            return False

    sig = ECDSASignature.fromSignature(sigBytes[lenHash:])
    try:
        pub = ecRecoverPub(hashedData, sig, chainId)
    except ValueError:
        return False

    addr = bytesToHex(pubToAddress(pub[1:]))
    return areSameAddresses(address, addr)

def isValidPrivate(privateKey: bytes) -> bool:
    """Checks whether the `privateKey` is a valid private key
    (follows the rules of the curve `secp256k1`).

    Args:
        privateKey (bytes): The private key of the AIN blockchain account that you want to check.

    Returns:
        bool: `True`, if the given key is a valid private key.
            `False`, if not.
    """
    try:
        if len(privateKey) != 32:
            return False
        validate_secret(privateKey)
        return True
    except:
        return False

def isValidPublic(publicKey: bytes, isSEC1: bool = False) -> bool:
    """Checks whether the `publicKey` is a valid public key
    (follows the rules of the curve `secp256k1` and meets the AIN requirements).

    Args:
        publicKey (bytes): The public key of the AIN blockchain account that you want to check.
        isSEC1 (bool): If `True`, public key should be SEC1 encoded one.
            Defaults to `False`.
    
    Returns:
        bool: `True`, if the given key is a valid public key.
            `False`, if not.
    """
    try:
        if len(publicKey) == 64:
            publicKey = bytes([4]) + publicKey
        else:
            if not isSEC1:
                return False
        pk = PublicKey(publicKey)
    except ValueError:
        return False
    return True

def areSameAddresses(address1: str, address2: str) -> bool:
    """Checks if the two addresses are the same.
    
    Args:
        address1 (str): The first address.
        address2 (str): The second address.
    
    Returns:
        bool: `True`, if the two addresses are the same.
            `False`, if not.
    """
    return toChecksumAddress(address1) == toChecksumAddress(address2)

def privateToPublic(privateKey: bytes) -> bytes:
    """Returns the public key of a given private key.
    
    Args:
        privateKey (bytes): The private key of the AIN blockchain account.

    Returns:
        bytes:
            The public key of a given private key.
    """
    prK = PrivateKey(privateKey)
    return prK.public_key.format(False)[1:]

def pubToAddress(publicKey: Union[bytes, str], isSEC1: bool = False) -> bytes:
    """Returns the AIN blockchain address of the given public key,
    which is the lower 160 bits of the Keccak-256 hash of the public key.

    Args:
        publicKey (Union[bytes, str]): The public key of the AIN blockchain account or SEC1 encoded public key.
            When the type is `str`, it must be hex prefixed.
        isSEC1 (bool): If `True`, public key should be SEC1 encoded one.
            Defaults to `False`.

    Returns:
        bytes: The AIN blockchain address of the given public key.
    """
    if type(publicKey) is str:
        publicKeyBytes = toBytes(publicKey)
    elif type(publicKey) is bytes:
        publicKeyBytes = publicKey
    else:
        raise ValueError("Invalid public key type")

    if isSEC1 and len(publicKeyBytes) != 64:
        publicKeyBytes = PublicKey(publicKeyBytes).format(False)[1:]
    if len(publicKeyBytes) != 64:
        raise ValueError("Invalid public key")
    return keccak(publicKeyBytes)[-20:]

def privateToAddress(privateKey: bytes) -> str:
    """Returns the checksummed AIN blockchain address of the given private key.

    Args:
        privateKey (bytes): The private key of an AIN blockchain account.

    Returns:
        str: The checksummed AIN blockchain address of the given private key.
    """
    return toChecksumAddress(bytesToHex(pubToAddress(privateToPublic(privateKey))))

def generateMnemonic() -> str:
    """Generates the random english mnemonic.

    Returns:
        str: The randomly generated english mnemonic.
    """
    return Mnemonic("english").generate()

def mnemonicToPrivatekey(mnemonic: str, index: int = 0) -> bytes:
    """Returns an private key with the given mnemonic.

    Args:
        mnemonic (str): The mnemonic of account.
        index (int): The index of account. Defaults to 0.

    Returns:
        bytes: The private key with the given mnemonic.
    """
    if index < 0:
        raise ValueError("index should be greater than 0")

    if not Mnemonic("english").check(mnemonic):
        raise ValueError("Invalid mnemonic")
    
    seed = Mnemonic.to_seed(mnemonic)
    bip32 = BIP32.from_seed(seed)
    path = AIN_HD_DERIVATION_PATH + f"{index}"
    return bip32.get_privkey_from_path(path)

# NOTE(kriii): Referenced https://github.com/bitchan/eccrypto/blob/master/index.js#L195-L258
def encryptWithPublicKey(publicKey: Union[bytes, str], message: str) -> ECIESEncrypted:
    """Encrypts message with publicKey.

    Args:
        publicKey (Union[bytes, str]): The public key.
            When the type is `str`, it must be hex prefixed.
        message (str): The message.

    Returns:
        ECIESEncrypted: The object of the encrypted message.
    """
    if type(publicKey) is str:
        publicKeyBytes = bytes.fromhex(publicKey)
    elif type(publicKey) is bytes:
        publicKeyBytes = publicKey
    else:
        raise ValueError("Invalid public key type")

    if len(publicKeyBytes) == 64:
        publicKeyBytes = bytes([4]) + publicKeyBytes

    pbK = PublicKey(publicKeyBytes)
    epK = PrivateKey()
    ephemPublicKey = epK.public_key.format(False)
    shared = pbK.multiply(epK.secret)
    derive = shared.format(True)[1:]
    hash = SHA512.new()
    hash.update(derive)
    iv = token_bytes(16)
    encKey = hash.digest()[0:32]
    aes = AES.new(encKey, AES.MODE_CBC, iv=iv)
    ciphertext = aes.encrypt(pad(toBytes(message), 16))
    macKey = hash.digest()[32:]
    hmac = HMAC.new(macKey, digestmod=SHA256)
    hmac.update(iv + ephemPublicKey + ciphertext)
    mac = hmac.digest()

    return ECIESEncrypted(iv, ephemPublicKey, ciphertext, mac)

def decryptWithPrivateKey(privateKey: Union[bytes, str], encrypted: ECIESEncrypted) -> str:
    """Decrypts encrypted data with privateKey.

    Args:
        privateKey (Union[bytes, str]): The private key.
        encrypted (ECIESEncrypted): The ECIES encrypted object.
    
    Returns:
        str: The decrypted message.
    """
    if type(privateKey) is str:
        privateKeyBytes = bytes.fromhex(privateKey)
    elif type(privateKey) is bytes:
        privateKeyBytes = privateKey
    else:
        raise ValueError("Invalid public key type")

    prK = PrivateKey(privateKeyBytes)
    epK = PublicKey(encrypted.ephemPublicKey)
    shared = epK.multiply(prK.secret)
    derive = shared.format(True)[1:]
    hash = SHA512.new()
    hash.update(derive)
    encKey = hash.digest()[0:32]
    macKey = hash.digest()[32:]
    hmac = HMAC.new(macKey, digestmod=SHA256)
    hmac.update(encrypted.iv + encrypted.ephemPublicKey + encrypted.ciphertext)
    mac = hmac.digest()

    if mac != encrypted.mac:
        raise ValueError("Bad MAC")

    aes = AES.new(encKey, AES.MODE_CBC, iv=encrypted.iv)
    plaintext = unpad(aes.decrypt(encrypted.ciphertext), 16)
    return plaintext.decode("utf-8")



import json
import uuid as UUID
from types import SimpleNamespace
from typing import Union, Optional
from secrets import token_bytes
from Crypto.Cipher import AES
from Crypto.Protocol.KDF import PBKDF2, scrypt
from Crypto.Hash import SHA256
from Crypto.Util.Padding import pad, unpad
from ain.utils import privateToAddress, keccak

class Cipherparams:
    """Class for the cipher params."""

    iv: str
    """The initial vector of the cipher."""

class Crypto:
    """Class for the crypto."""
    
    ciphertext: str
    """The ciphertext."""
    cipherparams: Cipherparams
    """The cipher params."""
    cipher: str
    """The name of the cipher."""
    kdf: str
    """The name of the KDF(key derivation function). `scrypt` and `pbkdf2` are supported."""
    kdfparams: dict
    """The KDF params."""
    mac: str
    """The MAC(message authentication code)."""

class V3KeystoreOptions:
    """Class for the v3 keystore options."""

    cipher: Optional[str]
    """The name of the cipher."""
    iv: Optional[bytes]
    """The initial vector of the cipher."""
    kdf: Optional[str]
    """The name of the KDF(key derivation function). `scrypt` and `pbkdf2` are supported."""
    salt: Optional[bytes]
    """The salt for the KDF."""
    dklen: Optional[int]
    """The length of the desired key for KDF."""
    n: Optional[int]
    """The n option for the KDF `scrypt`"""
    r: Optional[int]
    """The r option for the KDF `scrypt`"""
    p: Optional[int]
    """The p option for the KDF `scrypt`"""
    c: Optional[int]
    """The c option for the KDF `pbkdf2`"""
    uuid: Optional[bytes]
    """The predetermined bytes of the UUID."""

    def __init__(
        self,
        cipher: str = None,
        iv: bytes = None,
        kdf: str = None,
        salt: bytes = None,
        dklen: int = None,
        n: int = None,
        r: int = None,
        p: int = None,
        c: int = None,
        uuid: bytes = None,
    ):
        if cipher is not None:
            self.cipher = cipher
        if iv is not None:
            self.iv = iv
        if kdf is not None:
            self.kdf = kdf
        if salt is not None:
            self.salt = salt
        if dklen is not None:
            self.dklen = dklen
        if n is not None:
            self.n = n
        if r is not None:
            self.r = r
        if p is not None:
            self.r = p
        if c is not None:
            self.c = c
        if uuid is not None:
            self.uuid = uuid

class V3Keystore:
    """Class for the v3 keystore."""

    version: int
    """The version of the keystore."""
    id: str
    """The UUID of the keystore."""
    address: str
    """The AIN blockchain address of the keystore."""
    crypto: Crypto
    """The object of the crypto."""

    def __init__(
        self,
        version: int,
        id: str,
        address: str,
        ciphertext: str,
        iv: str,
        cipher: str,
        kdf: str,
        kdfparams: dict,
        mac: str
    ):
        self.version = version
        self.id = id
        self.address = address
        self.crypto = Crypto()
        self.crypto.ciphertext = ciphertext
        self.crypto.cipherparams = Cipherparams()
        self.crypto.cipherparams.iv = iv
        self.crypto.cipher = cipher
        self.crypto.kdf = kdf
        self.crypto.kdfparams = kdfparams
        self.crypto.mac = mac

    @classmethod
    def fromPrivateKey(
        cls,
        privateKey: Union[bytes, str],
        password: str,
        options: V3KeystoreOptions = V3KeystoreOptions()
    ):
        """Converts an account into a v3 Keystore and encrypts it with a password.
        
        Args:
            privateKey (Union[bytes, str]): The private key of an AIN blockchain account.
            password (str): The password of the v3 keystore.
            options (V3KeystoreOptions): The options for the v3 keystore.
                Defaults to no options.
        """

        if type(privateKey) is str:
            privateKeyBytes = bytes.fromhex(privateKey)
        elif type(privateKey) is bytes:
            privateKeyBytes = privateKey
        else:
            raise ValueError("Invalid public key type")

        salt = getattr(options, "salt", token_bytes(32))
        iv = getattr(options, "iv", token_bytes(16))
        uuid = getattr(options, "uuid", token_bytes(16))
        dklen = getattr(options, "dklen", 32)
        kdf = getattr(options, "kdf", "scrypt")

        kdfparams = {
            "dklen": dklen,
            "salt": salt.hex(),
        }
        if kdf == "scrypt":
            n = getattr(options, "n", 262144)
            r = getattr(options, "r", 8)
            p = getattr(options, "p", 1)
            kdfparams.update({
                "n": n,
                "r": r,
                "p": p,
            })
            derivedKeyTemp = scrypt(
                password,
                salt, # type: ignore
                dklen,
                N=n,
                r=r,
                p=p
            )
            if type(derivedKeyTemp) is bytes:
                derivedKey = derivedKeyTemp
            else:
                raise ValueError("scrypt derived multiple keys")
        elif kdf == "pbkdf2":
            c = getattr(options, "c", 262144)
            kdfparams.update({
                "c": c,
                "prf": "hmac-sha256",
            })
            derivedKey = PBKDF2(
                password,
                salt,
                dklen,
                count=c,
                hmac_hash_module=SHA256
            )
        else:
            raise ValueError("Unsupported kdf")

        cipherkey = derivedKey[0:16]
        padding = False
        cipher = getattr(options, "cipher", "aes-128-ctr")
        if cipher == "aes-128-ecb":
            aescipher = AES.new(cipherkey, AES.MODE_ECB)
            padding = True
        elif cipher == "aes-128-cbc" or cipher == "aes128":
            aescipher = AES.new(cipherkey, AES.MODE_CBC, iv)
            padding = True
        elif cipher == "aes-128-cfb":
            aescipher = AES.new(cipherkey, AES.MODE_CFB, iv, segment_size=128)
        elif cipher == "aes-128-cfb8":
            aescipher = AES.new(cipherkey, AES.MODE_CFB, iv)
        elif cipher == "aes-128-ofb":
            aescipher = AES.new(cipherkey, AES.MODE_OFB, iv)
        elif cipher == "aes-128-ctr":
            aescipher = AES.new(cipherkey, AES.MODE_CTR, nonce=iv[:8], initial_value=iv[8:])
        else:
            raise ValueError("Unsupported cipher")

        if padding:
            ciphertext = aescipher.encrypt(pad(privateKeyBytes, AES.block_size))
        else:
            ciphertext = aescipher.encrypt(privateKeyBytes)

        return cls(
            version=3,
            id=str(UUID.UUID(bytes=uuid, version=4)),
            address=privateToAddress(privateKeyBytes).lower().replace("0x",""),
            ciphertext=ciphertext.hex(),
            iv=iv.hex(),
            cipher=cipher,
            kdf=kdf,
            kdfparams=kdfparams,
            mac=keccak(derivedKey[16:32] + ciphertext).hex()
        )

    @classmethod
    def fromJSON(cls, v3KeystoreJSON: str):
        """Converts the stringified v3 keystore JSON into a v3 keystore object.
        
        Args:
            v3KeystoreJSON (str): The stringified v3 keystore JSON.
        """

        loaded = json.loads(v3KeystoreJSON, object_hook=lambda d: SimpleNamespace(**d))
        version = int(loaded.version)
        id = str(loaded.id)
        address = str(loaded.address)
        ciphertext = str(loaded.crypto.ciphertext)
        iv = str(loaded.crypto.cipherparams.iv)
        cipher = str(loaded.crypto.cipher)
        kdf = str(loaded.crypto.kdf)
        kdfparams = dict(loaded.crypto.kdfparams.__dict__)
        mac = str(loaded.crypto.mac)
        return cls(
            version,
            id,
            address,
            ciphertext,
            iv,
            cipher,
            kdf,
            kdfparams,
            mac
        )

    def toPrivateKey(self, password: str) -> bytes:
        """Returns a private key from a v3 Keystore."""

        kdfparams = self.crypto.kdfparams
        salt = bytes.fromhex(kdfparams["salt"])
        dklen = kdfparams["dklen"]
        if self.crypto.kdf == "scrypt":
            derivedKeyTemp = scrypt(
                password,
                salt, # type: ignore
                dklen,
                N=kdfparams["n"],
                r=kdfparams["r"],
                p=kdfparams["p"]
            )
            if type(derivedKeyTemp) is bytes:
                derivedKey = derivedKeyTemp
            else:
                raise ValueError("scrypt derived multiple keys")
        elif self.crypto.kdf == "pbkdf2":
            derivedKey = PBKDF2(
                password,
                salt,
                dklen,
                count=kdfparams["c"],
                hmac_hash_module=SHA256
            )
        else:
            raise ValueError("Unsupported kdf")

        ciphertext = bytes.fromhex(self.crypto.ciphertext)
        mac = keccak(derivedKey[16:32] + ciphertext).hex()
        if mac != self.crypto.mac:
            raise ValueError("Key derivation failed - possibly wrong password")
            
        cipherkey = derivedKey[0:16]
        padding = False
        cipher = self.crypto.cipher
        iv = bytes.fromhex(self.crypto.cipherparams.iv)
        if cipher == "aes-128-ecb":
            aescipher = AES.new(cipherkey, AES.MODE_ECB)
            padding = True
        elif cipher == "aes-128-cbc" or cipher == "aes128":
            aescipher = AES.new(cipherkey, AES.MODE_CBC, iv)
            padding = True
        elif cipher == "aes-128-cfb":
            aescipher = AES.new(cipherkey, AES.MODE_CFB, iv, segment_size=128)
        elif cipher == "aes-128-cfb8":
            aescipher = AES.new(cipherkey, AES.MODE_CFB, iv)
        elif cipher == "aes-128-ofb":
            aescipher = AES.new(cipherkey, AES.MODE_OFB, iv)
        elif cipher == "aes-128-ctr":
            aescipher = AES.new(cipherkey, AES.MODE_CTR, nonce=iv[:8], initial_value=iv[8:])
        else:
            raise ValueError("Unsupported cipher")

        if padding:
            privateKeyBytes = unpad(aescipher.decrypt(ciphertext), AES.block_size)
        else:
            privateKeyBytes = aescipher.decrypt(ciphertext)
        return privateKeyBytes

    def __str__(self):
        return json.dumps(
            self.__dict__,
            default=lambda o: o.__dict__,
            separators=(",", ":")
        )


from ain.provider import Provider
from ain.constants import BLOCKCHAIN_PROTOCOL_VERSION

class Network:
    """Class for the AIN Blockchain network status checker."""

    provider: Provider
    """The `Provider` instance."""
    protoVer: str
    """The protocol version of the blockchain."""
    
    def __init__(self, provider: Provider):
        self.provider = provider
        self.protoVer = BLOCKCHAIN_PROTOCOL_VERSION

    async def isListening(self):
        """Returns whether the node is listening for network connections."""
        return await self.provider.send("net_listening")
    
    async def getNetworkId(self):
        """Returns the id of the network the node is connected to."""
        return await self.provider.send("net_getNetworkId")

    async def checkProtocolVersion(self):
        """Checks the protocol version."""
        return await self.provider.send("ain_checkProtocolVersion")
    
    async def getProtocolVersion(self):
        """Returns the protocol version of the node."""
        return await self.provider.send("ain_getProtocolVersion")
    
    async def getPeerCount(self):
        """Returns the number of peers the provider node is connected to."""
        return await self.provider.send("net_peerCount")

    async def isSyncing(self):
        """Returns whether the node is syncing with the network or not."""
        return await self.provider.send("net_syncing")

    async def getEventHandlerNetworkInfo(self):
        """Returns the event handler network information."""
        return await self.provider.send("net_getEventHandlerNetworkInfo")



import json
import platform
import re
from unittest import TestCase
from snapshottest import TestCase as SnapshotTestCase
from ain.ain import Ain
from ain.provider import JSON_RPC_ENDPOINT
from ain.utils import *
from ain.utils.v3keystore import *
from ain.errors import *
from ain.types import *
from .data import (
    testNode,
    accountSk,
)
from .util import (
    asyncTest,
    getRequest,
    postRequest,
    waitUntilTxFinalized,
    eraseProtoVer,
)

TX_PATTERN = re.compile("0x[0-9a-fA-F]{64}")
PY_VERSION = platform.python_version().replace(".", "")
APP_NAME = "bfan_raw"

class TestCoreRaw(TestCase):
    ain: Ain
    defaultAddr: str
    addr1: str
    addr2: str
    
    @classmethod
    @asyncTest
    async def setUpClass(cls):
        cls.ain = Ain(testNode, 0, AinOptions(rawResultMode=True))
        cls.ain.wallet.addAndSetDefaultAccount(accountSk)
        newAccounts = cls.ain.wallet.create(2)
        cls.defaultAddr = cls.ain.wallet.defaultAccount.address
        cls.addr1 = newAccounts[0]
        cls.addr2 = newAccounts[1]
        getAddr = await getRequest(f"{testNode}/get_address")
        nodeAddr = getAddr["result"]
        stakeForApps = await postRequest(f"{testNode}/set", {
            "op_list": [
                {
                    "type": "SET_VALUE",
                    "ref": f"/staking/test_raw{PY_VERSION}/{nodeAddr}/0/stake/{getTimestamp()}/value",
                    "value": 1
                },
                {
                    "type": "SET_VALUE",
                    "ref": f"/staking/{APP_NAME}{PY_VERSION}/{nodeAddr}/0/stake/{getTimestamp()}/value",
                    "value": 1
                }
            ],
            "nonce": -1
        })
        await waitUntilTxFinalized(testNode, stakeForApps["result"]["tx_hash"])

        createApps = await postRequest(f"{testNode}/set", {
            "op_list": [
                {
                    "type": "SET_VALUE",
                    "ref": f"/manage_app/test_raw{PY_VERSION}/create/{getTimestamp()}",
                    "value": { "admin": { cls.defaultAddr: True } }
                },
                {
                    "type": "SET_VALUE",
                    "ref": f"/manage_app/{APP_NAME}{PY_VERSION}/create/{getTimestamp()}",
                    "value": { "admin": { cls.defaultAddr: True } }
                }
            ],
            "nonce": -1
        })
        await waitUntilTxFinalized(testNode, createApps["result"]["tx_hash"])
    
    @asyncTest
    async def test00ValidateAppNameTrueRaw(self):
        res = await self.ain.validateAppName("test_new")
        self.assertEqual(res["is_valid"], True)
        self.assertEqual(res["result"], True)
        self.assertEqual(res["code"], 0)
        self.assertTrue("protoVer" in res)

    @asyncTest
    async def test00ValidateAppNameFalseRaw(self):
        res = await self.ain.validateAppName("app/path")
        self.assertEqual(res["is_valid"], False)
        self.assertEqual(res["result"], False)
        self.assertEqual(res["code"], 30601)
        self.assertEqual(res["message"], "Invalid app name for state label: app/path")
        self.assertTrue("protoVer" in res)

    @asyncTest
    async def test00SendSignedTransactionRaw(self):
        tx = TransactionBody(
            nonce=-1,
            gas_price=500,
            timestamp=getTimestamp(),
            operation=SetOperation(
                type="SET_OWNER",
                ref=f"/apps/{APP_NAME}{PY_VERSION}",
                value={
                    ".owner": {
                        "owners": {
                            "*": {
                                "write_owner": True,
                                "write_rule": True,
                                "write_function": True,
                                "branch_owner": True
                            }
                        }
                    }
                }
            )
        )

        sig = self.ain.wallet.signTransaction(tx)
        res = await self.ain.sendSignedTransaction(sig, tx)
        targetTxHash = res["result"]["tx_hash"]
        self.assertFalse("code" in res)
        self.assertTrue(TX_PATTERN.fullmatch(targetTxHash) is not None)
        self.assertEqual(res["result"]["result"]["code"], 0)

    @asyncTest
    async def test00SendSignedTransactionInvalidSignatureRaw(self):
        tx = TransactionBody(
            nonce=-1,
            gas_price=500,
            timestamp=getTimestamp(),
            operation=SetOperation(
                type="SET_OWNER",
                ref=f"/apps/{APP_NAME}{PY_VERSION}",
                value={
                    ".owner": {
                        "owners": {
                            "*": {
                                "write_owner": True,
                                "write_rule": True,
                                "write_function": True,
                                "branch_owner": True
                            }
                        }
                    }
                }
            )
        )

        sig = ""
        res = await self.ain.sendSignedTransaction(sig, tx)
        self.assertEqual(res["result"], None)
        self.assertEqual(res["code"], 30302)
        self.assertEqual(res["message"], "Missing properties.")
        self.assertTrue("protoVer" in res)

    @asyncTest
    async def test01SendTransactionBatchRaw(self):
        tx1 = TransactionInput(
            operation=SetOperation(
                type="SET_VALUE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users",
                value={ self.defaultAddr: True }
            ),
            address=self.addr1
        )

        tx2 = TransactionInput(
            operation=SetMultiOperation(
                type="SET",
                op_list=[
                    SetOperation(
                        type="SET_OWNER",
                        ref=f"/apps/{APP_NAME}{PY_VERSION}/users",
                        value={
                            ".owner": {
                                "owners": {
                                    "*": {
                                        "write_owner": False,
                                        "write_rule": False,
                                        "write_function": True,
                                        "branch_owner": True
                                    }
                                }
                            }
                        }
                    ),
                    SetOperation(
                        type="SET_OWNER",
                        ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                        value={
                            ".owner": {
                                "owners": {
                                    self.defaultAddr: {
                                        "write_owner": True,
                                        "write_rule": True,
                                        "write_function": True,
                                        "branch_owner": True
                                    }
                                }
                            }
                        }
                    )
                ]
            ),
            address=self.addr2
        )

        tx3 = TransactionInput(
            operation=SetOperation(
                type="SET_RULE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                value={ ".rule": { "write": f'auth.addr === "{self.defaultAddr}"' } }
            )
        )

        tx4 = TransactionInput(
            operation=SetOperation(
                type="SET_VALUE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                value=False
            )
        )

        tx5 = TransactionInput(
            operation=SetOperation(
                type="SET_VALUE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                value=True
            ),
            address=self.addr1
        )

        tx6 = TransactionInput(
            operation=SetOperation(
                type="SET_RULE",
                ref=f"/apps/{APP_NAME}{PY_VERSION}/users/{self.defaultAddr}",
                value={ ".rule": { "write": "true" } }
            ),
            address=self.addr2
        )

        raw = await self.ain.sendTransactionBatch([tx1, tx2, tx3, tx4, tx5, tx6])
        resps = raw["result"]
        for res in resps:
            self.assertTrue(TX_PATTERN.fullmatch(res["tx_hash"]) is not None)
        self.assertEqual(resps[0]["result"]["code"], 12103)
        self.assertEqual(resps[1]["result"]["result_list"]["0"]["code"], 0)
        self.assertEqual(resps[1]["result"]["result_list"]["1"]["code"], 0)
        self.assertEqual(resps[2]["result"]["code"], 0)
        self.assertEqual(resps[3]["result"]["code"], 0)
        self.assertEqual(resps[4]["result"]["code"], 12103)
        self.assertEqual(resps[5]["result"]["code"], 12302)

    @asyncTest
    async def test01SendTransactionBatchWithEmptyListRaw(self):
        res = await self.ain.sendTransactionBatch([])
        self.assertEqual(res["result"], None)
        self.assertEqual(res["code"], 30401)
        self.assertEqual(res["message"], "Invalid batch transaction format.")
        self.assertTrue("protoVer" in res)

class TestDatabaseRaw(SnapshotTestCase):
    ain: Ain
    defaultAddr: str
    allowedPath: str

    @classmethod
    def setUpClass(cls):
        cls.ain = Ain(testNode, 0, AinOptions(rawResultMode=True))
        cls.ain.wallet.addAndSetDefaultAccount(accountSk)
        cls.defaultAddr = cls.ain.wallet.defaultAccount.address
        cls.allowedPath = f"apps/{APP_NAME}{PY_VERSION}/users/{cls.defaultAddr}"
        super(TestDatabaseRaw, cls).setUpClass()

    def matchSnapshot(self, value):
        dump = json.dumps(
            value,
            default=lambda o: o.__dict__,
            separators=(",", ":"),
            sort_keys=True
        ).replace(f'{APP_NAME}{PY_VERSION}', APP_NAME)
        self.assertMatchSnapshot(json.loads(dump))

    def testRef(self):
        self.assertEqual(self.ain.db.ref().path, "/")
        self.assertEqual(self.ain.db.ref(self.allowedPath).path, "/" + self.allowedPath)
    
    @asyncTest
    async def test01GetValue(self):
        res = eraseProtoVer(await self.ain.db.ref(self.allowedPath).getValue())
        self.matchSnapshot(res)
    
    @asyncTest
    async def test01Get(self):
        res = eraseProtoVer(await self.ain.db.ref(self.allowedPath).get([
            GetOperation(
                type="GET_RULE",
                ref=""
            ),
            GetOperation(
                type="GET_VALUE"
            ),
            GetOperation(
                type="GET_VALUE",
                ref="deeper/path"
            )
        ]))
        self.matchSnapshot(res)

    @asyncTest
    async def test01GetWithEmptyList(self):
        res = eraseProtoVer(await self.ain.db.ref(self.allowedPath).get([]))
        self.matchSnapshot(res)
